@@ -1,23 +1,38 @@
-/*             This file is part of the New World OS project
+/*
+--             This file is part of the New World OS project
 --                 Copyright (C) 2005-2007  QRW Software
 --           J. Scott Edwards - j.scott.edwards.nwos@gmail.com 
 --                      http://www.qrwsoftware.com
 --                      http://nwos.sourceforge.com
 --
--- NWOS is free software;  you can redistribute it and/or modify it under the
--- terms of the GNU General Public License  as published by the Free Software
--- Foundation; either version 2, or (at your option) any later version.  This
--- software is distributed with the hope that it will be useful,  but WITHOUT
--- ANY WARRANTY;  without  even the  implied warranty  of MERCHANTABILITY  or
--- FITNESS FOR A PARTICULAR PURPOSE.   See the GNU General Public License for
--- more  details.  You should have received a copy  of the GNU General Public
--- License along with this package;  see the file LICENSE.  If not, write to:
+--   This program is free software: you can redistribute it and/or modify
+--   it under the terms of the GNU General Public License as published by
+--   the Free Software Foundation, either version 3 of the License, or
+--   (at your option) any later version.
+--
+--   This program is distributed in the hope that it will be useful,
+--   but WITHOUT ANY WARRANTY; without even the implied warranty of
+--   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+--   GNU General Public License for more details.
+--
+--   You should have received a copy of the GNU General Public License
+--   along with this program, in the file LICENSE.  If not, see 
+--   <http://www.gnu.org/licenses/>.
+--
+--   You can also contact me via paper mail at:
+--
+--      QRW Software
+--      P.O. Box 27511
+--      Salt Lake City, UT 84127-0511, USA.
 --
---      Free Software Foundation, Inc.
---      59 Temple Place - Suite 330
---      Boston, MA 02111-1307, USA.
 --
 -- $Log: computer.c,v $
+-- Revision 1.28  2007/07/12 03:12:42  jsedwards
+-- Add a command to add a birthday to a person.
+--
+-- Revision 1.27  2007/07/01 19:44:11  jsedwards
+-- Upgrade to GPLv3.
+--
 -- Revision 1.26  2007/05/24 13:27:13  jsedwards
 -- Added list all manufacturers command.
 --
@@ -296,6 +311,22 @@ void execute_command(char *command)
 	ObjRef  city_ref;
 	nwos_add_city(&city_ref);
     }
+    else if (strstr(command, " birthday is ") != NULL)
+    {
+	ptr1 = strstr(command, " birthday is ");
+	ptr2 = ptr1 + strlen(" birthday is ");      /* point to the start of the phone number */
+	if (ptr1[-2] == '\'' && ptr1[-1] == 's')
+	{
+	    ptr1 -= 2;
+	}
+	else if (ptr1[-1] == '\'')
+	{
+	    ptr1 -= 1;
+	}
+	*ptr1 = '\0';    /* terminate the name here */
+
+	nwos_add_birthday(command, ptr2);
+    }
     else if (strstr(command, " mobile phone is ") != NULL)
     {
 	ptr1 = strstr(command, " mobile phone is ");
@@ -372,6 +403,7 @@ int main(int argc, char* argv[])
 		printf("   \"add city\" - adds a new city to the database (not of much use in this version).\n");
 		printf("   \"describe <name>\" - describes any person(s) with the name <name> in the database.\n");
 		printf("   \"list cities in <state>\" - lists the cities in the database for state <state>.\n");
+		printf("   \"<name>'s birthday is <date>\" - adds the birthday <date> to the person <name>.\n");
 		printf("   \"<name>'s mobile phone is <number>\" - adds the phone number <number> to the person <name>.\n");
 		printf("   \"what is <name>'s number\" - prints out the phone number for the person <name>.\n");
 		printf("   \"list all credit unions\" - list all credit unions in the database.\n");
