@@ -1,23 +1,53 @@
-/*             This file is part of the New World OS project
+/*
+--             This file is part of the New World OS project
 --                Copyright (C) 2006, 2007  QRW Software
 --           J. Scott Edwards - j.scott.edwards.nwos@gmail.com 
 --                      http://www.qrwsoftware.com
 --                      http://nwos.sourceforge.com
 --
--- NWOS is free software;  you can redistribute it and/or modify it under the
--- terms of the GNU General Public License  as published by the Free Software
--- Foundation; either version 2, or (at your option) any later version.  This
--- software is distributed with the hope that it will be useful,  but WITHOUT
--- ANY WARRANTY;  without  even the  implied warranty  of MERCHANTABILITY  or
--- FITNESS FOR A PARTICULAR PURPOSE.   See the GNU General Public License for
--- more  details.  You should have received a copy  of the GNU General Public
--- License along with this package;  see the file LICENSE.  If not, write to:
+--   This program is free software: you can redistribute it and/or modify
+--   it under the terms of the GNU General Public License as published by
+--   the Free Software Foundation, either version 3 of the License, or
+--   (at your option) any later version.
+--
+--   This program is distributed in the hope that it will be useful,
+--   but WITHOUT ANY WARRANTY; without even the implied warranty of
+--   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+--   GNU General Public License for more details.
+--
+--   You should have received a copy of the GNU General Public License
+--   along with this program, in the file LICENSE.  If not, see 
+--   <http://www.gnu.org/licenses/>.
+--
+--   You can also contact me via paper mail at:
+--
+--      QRW Software
+--      P.O. Box 27511
+--      Salt Lake City, UT 84127-0511, USA.
 --
---      Free Software Foundation, Inc.
---      59 Temple Place - Suite 330
---      Boston, MA 02111-1307, USA.
 --
 -- $Log: expand_sparse.c,v $
+-- Revision 1.21  2007/07/30 16:04:35  jsedwards
+-- Changed to allow either 0023 or 0024 compressed files (since they are
+-- identical) to be read in.
+--
+-- Revision 1.20  2007/07/03 01:21:46  jsedwards
+-- Changed to allow reading a 0023 compressed file into a 0024 storage device,
+-- to allow upgrading.
+--
+-- Revision 1.19  2007/07/01 19:44:11  jsedwards
+-- Upgrade to GPLv3.
+--
+-- Revision 1.18  2007/06/28 13:46:31  jsedwards
+-- Added assert to verify num_blocks stays in sync with nwos_private_blocks_used.
+--
+-- Revision 1.17  2007/06/28 13:44:56  jsedwards
+-- Changed so that final message about blocks used and total blocks mismatch
+-- are put in log as well as printed to the screen.
+--
+-- Revision 1.16  2007/06/26 13:55:23  jsedwards
+-- Changes for new (0023) format where chunks need to be allocated.
+--
 -- Revision 1.15  2007/06/20 22:47:45  jsedwards
 -- Add check to make sure disk is empty first, and if restoring from 0021
 -- compressed also fix the number of blocks used for bit maps when changing
@@ -115,8 +145,6 @@
 
 #include "objectify_private.h"
 
-/* this is a special function added to objectify.c, can be removed when no longer needed */
-extern void special_update_total_blocks_for_expand_0021(uint32 total_private_blocks);
 
 
 static void print_usage(char *program)
@@ -135,10 +163,15 @@ int main(int argc, char* argv[])
     int num_blocks = 0;
     uint32 total_private_blocks;
     uint32 used_private_blocks;
+    uint32 used_private_chunks;
     uint32 num_chunks;
     Disk_Header compressed_header;
+#if 0
     Disk_Header_0021 old_header;
+#endif
     char answer[8];
+    char msg[128];
+
 
     if (argc != 2)
     {
@@ -204,38 +237,20 @@ int main(int argc, char* argv[])
     /* verify that the number of used blocks is consistent with empty disk */
     num_chunks = (nwos_total_private_blocks + BLOCKS_IN_CHUNK - 1) / BLOCKS_IN_CHUNK; /* round up */
 
-    if (num_chunks * BIT_MAP_BLOCKS != nwos_used_private_blocks)
+    if (nwos_used_private_blocks != 0 || nwos_used_private_chunks != 0)
     {
-	fprintf(stderr, "ERROR: blocks used: %u does not match empty disk: %u!.\n",
-		nwos_used_private_blocks, num_chunks * BIT_MAP_BLOCKS);
+	fprintf(stderr, "ERROR: disk not empty, chunks used: %u  blocks used: %u!.\n", 
+		nwos_used_private_blocks, nwos_used_private_chunks);
 	exit(1);
     }
 
 
-    /* Allow a 0021 compressed file to be converted */
+    /* Version 0023, 0024 and 0025 archives are identical so allow expanding 0023 or 0024 */
 
-    if (memcmp(&block[4], "0021", 4) == 0)
+    if (memcmp(&block[4], "0023", 4) == 0 || memcmp(&block[4], "0024", 4) == 0)
     {
-	memcpy(&old_header, block, sizeof(old_header));
-
-	nwos_4_uint8_to_uint32(old_header.total_private_blocks, &total_private_blocks);
-	nwos_4_uint8_to_uint32(old_header.used_private_blocks, &used_private_blocks);
-
-	fprintf(stderr, "WARNING: compressed file is version 0021, converting.\n");
-	printf("WARNING: compressed file is version 0021, converting.\n");
-
-	if (total_private_blocks < nwos_total_private_blocks)
-	{
-	    fprintf(stderr, "Old disk size was %u and new is %u, using old!.\n",
-		    nwos_total_private_blocks, total_private_blocks);
-
-	    num_chunks = (total_private_blocks + BLOCKS_IN_CHUNK - 1) / BLOCKS_IN_CHUNK; /* round up */
-
-	    nwos_used_private_blocks = num_chunks * BIT_MAP_BLOCKS;
-
-	    /* remove this function from objectify.c when it is no longer needed */
-	    special_update_total_blocks_for_expand_0021(total_private_blocks);
-	}
+	fprintf(stderr, "WARNING: compressed file is version 002%c, is compatible.\n", block[7]);
+	printf("WARNING: compressed file is version 002%c, is compatible.\n", block[7]);
 
 	sleep(5);
     }
@@ -253,10 +268,14 @@ int main(int argc, char* argv[])
 	    exit(1);
 	}
 
+    }  /* remove these two braces if the the blocks and chunks */
+    {  /* need to be read differently for the old file format */ 
+
 	memcpy(&compressed_header, block, sizeof(compressed_header));
 
 	nwos_4_uint8_to_uint32(compressed_header.total_blocks, &total_private_blocks);
 	nwos_4_uint8_to_uint32(compressed_header.used_blocks, &used_private_blocks);
+	nwos_4_uint8_to_uint32(compressed_header.used_chunks, &used_private_chunks);
 
 	printf("compressed file version: %c%c%c%c\n",
 	       compressed_header.version_string[0],
@@ -268,10 +287,10 @@ int main(int argc, char* argv[])
     printf("   total_private_blocks: %u  used_private_blocks: %u\n",
 	   total_private_blocks, used_private_blocks);
 
-    if (total_private_blocks != nwos_total_private_blocks)
+    if (used_private_chunks > nwos_total_private_chunks)
     {
-	fprintf(stderr, "total_private_blocks mismatch, disk: %u  compressed %u\n",
-		nwos_total_private_blocks, total_private_blocks);
+	fprintf(stderr, "used chunks: %u greater than disk space: %u\n",
+		used_private_chunks, nwos_total_private_chunks);
 	exit(1);
     }
 
@@ -285,6 +304,15 @@ int main(int argc, char* argv[])
 
 	hash = nwos_hash_uint32_ref(ref);
 
+	if (hash == 0)
+	{
+	    nwos_allocate_new_chunk(ref);
+
+	    hash = nwos_hash_uint32_ref(ref);
+
+	    assert(hash != 0);
+	}
+
 	printf("id: %08x  block: %08x\n", ref, hash);
 	fflush(stdout);
 
@@ -300,6 +328,8 @@ int main(int argc, char* argv[])
 
 	num_blocks++;
 
+	assert(num_blocks == nwos_used_private_blocks);
+
 	bytes_read = fread(block, 1, FILE_BLOCK_SIZE, ifp);
     }
 
@@ -317,12 +347,16 @@ int main(int argc, char* argv[])
 
     nwos_terminate_objectify();
 
-    printf("Number of blocks written: %d\n", num_blocks);
+    snprintf(msg, sizeof(msg), "Number of blocks written: %d  chunks_used: %u",
+	     num_blocks, nwos_used_private_chunks);
+    nwos_log(msg);
+    printf("%s\n", msg);
 
     if (used_private_blocks != nwos_used_private_blocks)
     {
-	fprintf(stderr, "WARNING: used_private_blocks mismatch, disk: %u  compressed %u\n",
+	snprintf(msg, sizeof(msg), "WARNING: used_private_blocks mismatch, disk: %u  compressed %u\n",
 		nwos_used_private_blocks, used_private_blocks);
+	fprintf(stderr, "%s\n", msg);
     }
 
     return 0;
