@@ -1,25 +1,65 @@
-/*             This file is part of the New World OS project
+/*
+--             This file is part of the New World OS project
 --                 Copyright (C) 2006-2007  QRW Software
 --           J. Scott Edwards - j.scott.edwards.nwos@gmail.com 
 --                      http://www.qrwsoftware.com
 --                      http://nwos.sourceforge.com
 --
--- NWOS is free software;  you can redistribute it and/or modify it under the
--- terms of the GNU General Public License  as published by the Free Software
--- Foundation; either version 2, or (at your option) any later version.  This
--- software is distributed with the hope that it will be useful,  but WITHOUT
--- ANY WARRANTY;  without  even the  implied warranty  of MERCHANTABILITY  or
--- FITNESS FOR A PARTICULAR PURPOSE.   See the GNU General Public License for
--- more  details.  You should have received a copy  of the GNU General Public
--- License along with this package;  see the file LICENSE.  If not, write to:
+--   This program is free software: you can redistribute it and/or modify
+--   it under the terms of the GNU General Public License as published by
+--   the Free Software Foundation, either version 3 of the License, or
+--   (at your option) any later version.
+--
+--   This program is distributed in the hope that it will be useful,
+--   but WITHOUT ANY WARRANTY; without even the implied warranty of
+--   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+--   GNU General Public License for more details.
+--
+--   You should have received a copy of the GNU General Public License
+--   along with this program, in the file LICENSE.  If not, see 
+--   <http://www.gnu.org/licenses/>.
+--
+--   You can also contact me via paper mail at:
+--
+--      QRW Software
+--      P.O. Box 27511
+--      Salt Lake City, UT 84127-0511, USA.
 --
---      Free Software Foundation
---      51 Franklin Street, Fifth Floor
---      Boston, MA 02110-1301
---      USA
---      http://www.fsf.org/licenses
 --
 -- $Log: file.c,v $
+-- Revision 1.74  2007/07/30 18:49:31  jsedwards
+-- Changed to NOT attempt to create the updated File private class definition
+-- when an old object is updated.
+--
+-- Revision 1.73  2007/07/29 19:37:51  jsedwards
+-- Bug #1763165 - Fixed bug in nwos_find_matching_path_and_file_association
+-- where it would return ERROR_OCCURRED (non-zero) when an error occurred.
+-- Since the return value is boolean and ERROR_OCCURRED is non-zero, it
+-- appeared to return true even when an error occurred.
+--
+-- Revision 1.72  2007/07/28 04:01:35  jsedwards
+-- Converted to use uint64 for file size instead of uint32 and added function
+-- to convert an object in the old format to the new.
+--
+-- Revision 1.71  2007/07/24 16:57:11  jsedwards
+-- Modified nwos_number_of_files_for_path and nwos_file_path_to_path_and_file_
+-- association to skip over files that have a later revision.
+--
+-- Revision 1.70  2007/07/24 00:16:55  jsedwards
+-- Added nwos_add_new_revision_of_file function to create a new revision of a
+-- file that is already in the system.
+--
+-- Revision 1.69  2007/07/21 14:55:00  jsedwards
+-- Added 'match' parameter to nwos_file_is_identical so that it can ignore the
+-- time stamp on a file.
+--
+-- Revision 1.68  2007/07/05 17:59:07  jsedwards
+-- Change nwos_create_disc_copy to move disc to a new location if the disc
+-- (number) already exists and location is different.
+--
+-- Revision 1.67  2007/07/01 19:44:11  jsedwards
+-- Upgrade to GPLv3.
+--
 -- Revision 1.66  2007/06/18 13:16:17  jsedwards
 -- Fix so create_file_without_storing correctly adds file path separator.
 --
@@ -393,6 +433,15 @@ static size_t get_path_object_size(void* file_path_obj)
 }
 
 
+static inline uint64 file_size_to_uint64(uint8 size[5])
+{
+    return ((uint64)size[0] << 32) |
+	   ((uint64)size[1] << 24) |
+	   ((uint64)size[2] << 16) |
+	   ((uint64)size[3] << 8)  |
+	    (uint64)size[4];
+}
+
 
 /*-------------------------------------------------------------------------------------------------------------------*/
 /* File path (name) object stuff */
@@ -461,6 +510,8 @@ int nwos_number_of_files_for_path(ObjRef* path_ref)
 {
     uint8 kludge[FILE_BLOCK_SIZE];
     C_struct_File_Path* ptr_path_obj = (C_struct_File_Path*)kludge;
+    C_struct_Path_And_File_Association assoc_obj;
+    C_struct_File file_obj;
     ObjRef assoc_class_ref;
     ObjRef object_class;
     ReferenceList* ref_list;
@@ -483,7 +534,19 @@ int nwos_number_of_files_for_path(ObjRef* path_ref)
 	{
 	    nwos_get_object_class(&ref_list->references[i], &object_class);
 
-	    if (is_same_object(&object_class, &assoc_class_ref)) result++;
+	    if (is_same_object(&object_class, &assoc_class_ref))
+	    {
+		nwos_read_object_from_disk(&ref_list->references[i], &assoc_obj, sizeof(assoc_obj));
+
+		assert(is_same_object(&assoc_obj.path, path_ref));
+
+		nwos_read_object_from_disk(&assoc_obj.file, &file_obj, sizeof(file_obj));
+
+		if (is_void_reference(&file_obj.header.object.next_version))
+		{
+		    result++;
+		}
+	    }
 	}
 
 	nwos_free_reference_list(ref_list);
@@ -599,6 +662,7 @@ bool nwos_file_path_to_path_and_file_association(ObjRef* path_ref, int index, Ob
     ObjRef object_class;
     ObjRef assoc_class_ref;
     C_struct_Path_And_File_Association assoc_obj;
+    C_struct_File file_obj;
     int i;
     int count;
 
@@ -622,14 +686,23 @@ bool nwos_file_path_to_path_and_file_association(ObjRef* path_ref, int index, Ob
 	{
 	    nwos_read_object_from_disk(&ref_list->references[i], &assoc_obj, sizeof(assoc_obj));
 
-	    if (count == index)    /* this is the one we want */
+	    assert(is_same_object(&assoc_obj.path, path_ref));
+
+	    nwos_read_object_from_disk(&assoc_obj.file, &file_obj, sizeof(file_obj));
+
+	    /* don't count any that have newer versions */
+
+	    if (is_void_reference(&file_obj.header.object.next_version))
 	    {
-		copy_reference(assoc_ref, &ref_list->references[i]);
-		break;
-	    }
-	    else
-	    {
-		count++;
+		if (count == index)    /* this is the one we want */
+		{
+		    copy_reference(assoc_ref, &ref_list->references[i]);
+		    break;
+		}
+		else
+		{
+		    count++;
+		}
 	    }
 	}
     }
@@ -851,14 +924,14 @@ ObjCreateResult nwos_create_sha1(uint8 sha1sum[20], ObjRef* ref)
 /* File object stuff */
 /*-------------------------------------------------------------------------------------------------------------------*/
 
-static uint32 checksum_file(char* path, uint8 md5_digest[16], uint8 sha1_digest[20])
+static uint64 checksum_file(char* path, uint8 md5_digest[16], uint8 sha1_digest[20])
 {
     FILE* fp;
     MD5_CTX md5_context;    /* MD5 checksum context */
     struct sha1_ctx sha1_context;
     size_t bytes_read;
     uint8 buffer[4096];
-    uint32 file_length;
+    uint64 file_length;
 
     fp = fopen(path, "rb");
 
@@ -903,13 +976,13 @@ static uint32 checksum_file(char* path, uint8 md5_digest[16], uint8 sha1_digest[
 
 /* find a matching file from the md5sum */
 
-bool nwos_find_matching_file_from_size_md5_sha1(uint32 file_size, uint8 md5_digest[16], uint8 sha1_digest[20], ObjRef* ref)
+bool nwos_find_matching_file_from_size_md5_sha1(uint64 file_size, uint8 md5_digest[16], uint8 sha1_digest[20], ObjRef* ref)
 {
     ObjRef file_class_ref;
     ObjRef object_class;
     ObjRef md5sum_ref;
     ObjRef sha1sum_ref;
-    uint32 size;
+    uint64 size;
     C_struct_SHA1sum sha1_obj;
     C_struct_File file_obj;
     ReferenceList* ref_list;
@@ -945,10 +1018,9 @@ bool nwos_find_matching_file_from_size_md5_sha1(uint32 file_size, uint8 md5_dige
 
 		assert(is_same_object(&sha1sum_ref, &file_obj.sha1sum));
 
-		size = ((uint32)file_obj.size[0] << 24) | ((uint32)file_obj.size[1] << 16)
-		                      | ((uint32)file_obj.size[2] << 8) | (uint32)file_obj.size[3];
+		size = file_size_to_uint64(file_obj.size);
 
-		if (verbose) printf("size: %u\n", size);
+		if (verbose) printf("size: %llu\n", size);
 
 		if (size == file_size && is_same_object(&file_obj.md5sum, &md5sum_ref))
 		{
@@ -994,7 +1066,7 @@ ObjCreateResult nwos_create_file_without_storing_data(char* directory, char* fil
     ObjRef file_class_ref;
     ObjRef path_ref;
     ObjRef file_ref;
-    uint32 file_length;
+    uint64 file_length;
     C_struct_File file_obj;
     TimeStamp mod_time;
     uint8 md5_digest[16];
@@ -1022,9 +1094,9 @@ ObjCreateResult nwos_create_file_without_storing_data(char* directory, char* fil
 	return ERROR_OCCURRED;
     }
 
-    if (stat_struct.st_size > 2147483647)
+    if (stat_struct.st_size > 1099511627775)
     {
-	fprintf(stderr, "Cannot handle files larger than 2GB at this time\n");
+	fprintf(stderr, "Cannot handle files larger than 1 terabyte at this time\n");
 	return ERROR_OCCURRED;
     }
 
@@ -1038,7 +1110,7 @@ ObjCreateResult nwos_create_file_without_storing_data(char* directory, char* fil
 
     file_length = checksum_file(path, md5_digest, sha1_digest);
 
-    printf("length: %u\n", file_length);
+    printf("length: %llu\n", file_length);
 
     nwos_convert_timespec_to_time_stamp(&stat_struct.st_mtimespec, mod_time);
 
@@ -1066,10 +1138,11 @@ ObjCreateResult nwos_create_file_without_storing_data(char* directory, char* fil
 	nwos_fill_in_common_header(&file_obj.header.common, &file_ref, &file_class_ref);
 
 	/* store file size as 4 bytes in big endian order */
-	file_obj.size[0] = file_length >> 24;
-	file_obj.size[1] = file_length >> 16;
-	file_obj.size[2] = file_length >> 8;
-	file_obj.size[3] = file_length;
+	file_obj.size[0] = file_length >> 32;
+	file_obj.size[1] = file_length >> 24;
+	file_obj.size[2] = file_length >> 16;
+	file_obj.size[3] = file_length >> 8;
+	file_obj.size[4] = file_length;
 
 	nwos_create_md5(md5_digest, &file_obj.md5sum);
 
@@ -1096,11 +1169,11 @@ ObjCreateResult nwos_create_file_without_storing_data(char* directory, char* fil
 }
 
 
-static uint32 read_file(char* path, uint8 md5_digest[16], uint8 sha1_digest[20], ObjRef* block_list_ref)
+static uint64 read_file(char* path, uint8 md5_digest[16], uint8 sha1_digest[20], ObjRef* block_list_ref)
 {
     FILE* fp;
     size_t bytes_read;
-    uint32 file_length;
+    uint64 file_length;
     ObjRef data_block_ref;
     DataObject data_obj;
     MD5_CTX md5_context;    /* MD5 checksum context */
@@ -1176,7 +1249,7 @@ static uint32 read_file(char* path, uint8 md5_digest[16], uint8 sha1_digest[20],
 
     fclose(fp);
 
-    snprintf(msg, sizeof(msg), "length: %u", file_length);
+    snprintf(msg, sizeof(msg), "length: %llu", file_length);
     nwos_log(msg);
     printf("%s\n", msg);
 
@@ -1206,7 +1279,7 @@ ObjCreateResult nwos_create_file(char* path, ObjRef* ref)
     ObjRef file_class_ref;
     ObjRef path_ref;
     ObjRef file_ref;
-    uint32 file_length;
+    uint64 file_length;
     C_struct_File file_obj;
     uint8 md5_digest[16];
     uint8 sha1_digest[20];
@@ -1229,7 +1302,7 @@ ObjCreateResult nwos_create_file(char* path, ObjRef* ref)
 
     if (stat_struct.st_size > 1073741824)
     {
-	fprintf(stderr, "Cannot handle files larger than 1GB at this time\n");
+	fprintf(stderr, "Cannot store files larger than 1 gigabyte at this time\n");
 	return ERROR_OCCURRED;
     }
 
@@ -1296,10 +1369,11 @@ ObjCreateResult nwos_create_file(char* path, ObjRef* ref)
 	fflush(stdout);
 
 	/* store file size as 4 bytes in big endian order */
-	file_obj.size[0] = file_length >> 24;
-	file_obj.size[1] = file_length >> 16;
-	file_obj.size[2] = file_length >> 8;
-	file_obj.size[3] = file_length;
+	file_obj.size[0] = file_length >> 32;
+	file_obj.size[1] = file_length >> 24;
+	file_obj.size[2] = file_length >> 16;
+	file_obj.size[3] = file_length >> 8;
+	file_obj.size[4] = file_length;
 
 	nwos_create_reference_list(&file_ref, &file_obj.header.object.references);
 
@@ -1330,6 +1404,171 @@ ObjCreateResult nwos_create_file(char* path, ObjRef* ref)
 }
 
 
+/* Add new revision of existing file - returns DUPLICATE_FILE if an exact file match already existed */
+/*          (exact match meaning size, md5, sha1, date, and path are all identical).    */
+
+ObjCreateResult nwos_add_new_revision_of_file(char* path, ObjRef* ref)
+{
+    size_t length;
+    struct stat stat_struct;
+    ObjRef assoc_class_ref;
+    ObjRef path_ref;
+    ObjRef file_ref;
+    EveryObject header;
+    ReferenceList* ref_list;
+    int num_refs;
+    int i;
+    ObjRef object_class;
+    uint64 file_length;
+    C_struct_File old_file_obj;
+    C_struct_File new_file_obj;
+    C_struct_Path_And_File_Association assoc_obj;
+    uint8 md5_digest[16];
+    uint8 sha1_digest[20];
+    TimeStamp mod_time;
+
+    void_reference(ref);   /* in case it fails */
+
+
+    if (stat(path, &stat_struct) != 0)
+    {
+	perror(path);
+	return ERROR_OCCURRED;       /* FIX THIS to handle this more gracefully */
+    }
+
+    if (stat_struct.st_size == 0)
+    {
+	fprintf(stderr, "Cannot handle empty files at this time\n");
+	return ERROR_OCCURRED;
+    }
+
+    if (stat_struct.st_size > 1073741824)
+    {
+	fprintf(stderr, "Cannot handle files larger than 1 gigabyte at this time\n");
+	return ERROR_OCCURRED;
+    }
+
+    length = strlen(path);
+
+    assert(length <= 255);
+
+
+    /* first find out if we already have this file stored */
+
+    file_length = checksum_file(path, md5_digest, sha1_digest);
+
+    if (nwos_find_matching_file_from_size_md5_sha1(file_length, md5_digest, sha1_digest, &file_ref))
+    {
+	return DUPLICATE_FILE;
+    }
+
+
+    /* second find out if we already have this path */
+
+    if (!nwos_find_file_path(path, &path_ref))   /* this file must already exist */
+    {
+	return PATH_NOT_FOUND;
+    }
+
+    nwos_read_object_headers_from_disk(&path_ref, &header);
+
+    ref_list = nwos_malloc_reference_list(&header.object.references);
+    num_refs = ref_list->common_header.num_refs;
+
+    assert(nwos_find_private_class_definition("PATH AND FILE ASSOCIATION", &assoc_class_ref));
+
+    memset(&old_file_obj, 0, sizeof(old_file_obj));  /* zero it out */
+
+    for (i = 0; i < num_refs; i++)
+    {
+	nwos_get_object_class(&ref_list->references[i], &object_class);
+
+	if (is_same_object(&object_class, &assoc_class_ref))
+	{
+	    nwos_read_object_from_disk(&ref_list->references[i], &assoc_obj, sizeof(assoc_obj));
+
+	    assert(is_same_object(&assoc_obj.path, &path_ref));
+
+	    nwos_read_object_from_disk(&assoc_obj.file, &new_file_obj, sizeof(new_file_obj));
+
+	    if (is_void_reference(&new_file_obj.header.object.next_version))
+	    {
+		if (!is_void_reference(&old_file_obj.header.common.id))  /* already found a file */
+		{
+		    fprintf(stderr, "Cannot handle multiple files with the same name at this time\n");
+		    return MULTIPLE_FILES;
+		}
+
+		memcpy(&old_file_obj, &new_file_obj, sizeof(old_file_obj));
+	    }
+	}
+    }
+
+    printf("adding file revision: %s\n", path);
+
+    file_length = file_size_to_uint64(old_file_obj.size);
+
+    printf("previous file: %08x  length: %llu\n", nwos_ref_to_word(&old_file_obj.header.common.id), file_length);
+
+    nwos_generate_new_id(&file_ref);
+
+    copy_reference(&old_file_obj.header.object.next_version, &file_ref);
+
+    nwos_crc32_calculate((uint8*) &old_file_obj.header.object, sizeof(ObjectHeader), old_file_obj.header.common.header_chksum);
+
+    nwos_overwrite_object_to_disk(&old_file_obj.header.common.id, &old_file_obj, sizeof(old_file_obj));
+
+    printf("new file: %08x  ", nwos_ref_to_word(&file_ref));
+    fflush(stdout);
+
+    memset(&new_file_obj, 0, sizeof(new_file_obj));  /* zero it out */
+
+    nwos_fill_in_common_header(&new_file_obj.header.common, &file_ref, &old_file_obj.header.common.class_definition);
+
+    copy_reference(&new_file_obj.header.object.prev_version, &old_file_obj.header.common.id);
+
+    file_length = read_file(path, md5_digest, sha1_digest, &new_file_obj.block_list);
+
+    nwos_create_md5(md5_digest, &new_file_obj.md5sum);
+    nwos_create_sha1(sha1_digest, &new_file_obj.sha1sum);
+
+    nwos_add_to_references(&file_ref, &new_file_obj.md5sum);   /* add to this md5sum object's reference list */
+    nwos_add_to_references(&file_ref, &new_file_obj.sha1sum);   /* add to this md5sum object's reference list */
+
+      printf("MD5 sum: ");
+      for (i = 0; i < 16; i++) printf("%02x", md5_digest[i]);
+      printf("\n");
+      printf("SHA1 sum: ");
+      for (i = 0; i < 20; i++) printf("%02x", sha1_digest[i]);
+      printf("\n");
+
+    fflush(stdout);
+
+    /* store file size as 4 bytes in big endian order */
+    new_file_obj.size[0] = file_length >> 32;
+    new_file_obj.size[1] = file_length >> 24;
+    new_file_obj.size[2] = file_length >> 16;
+    new_file_obj.size[3] = file_length >> 8;
+    new_file_obj.size[4] = file_length;
+
+    nwos_create_reference_list(&file_ref, &new_file_obj.header.object.references);
+
+    nwos_crc32_calculate((uint8*) &new_file_obj.header.object, sizeof(ObjectHeader), new_file_obj.header.common.header_chksum);
+
+    nwos_crc32_calculate((uint8*) &new_file_obj.size[0], sizeof(C_struct_File) - sizeof(EveryObject), new_file_obj.header.common.data_chksum);
+
+    nwos_write_object_to_disk(&file_ref, &new_file_obj, sizeof(new_file_obj));
+
+    nwos_add_to_references(&file_ref, &new_file_obj.header.common.class_definition);
+
+    nwos_convert_timespec_to_time_stamp(&stat_struct.st_mtimespec, mod_time);
+
+    assert(nwos_find_or_create_path_and_file_association(&path_ref, &file_ref, mod_time, ref) == CREATED_NEW);
+
+    return CREATED_NEW_REVISION;
+}
+
+
 /* Check a file's md5sum */
 
 CheckFileResult nwos_check_file_checksums(char* path)
@@ -1341,7 +1580,7 @@ CheckFileResult nwos_check_file_checksums(char* path)
     ObjRef path_ref;
     ObjRef file_ref;
     uint8 buffer[4096];
-    uint32 file_length;
+    uint64 file_length;
     int num_blocks;
     C_struct_File file_obj;
     MD5_CTX md5_context;    /* MD5 checksum context */
@@ -1365,9 +1604,9 @@ CheckFileResult nwos_check_file_checksums(char* path)
 	return ERROR_OCCURRED;
     }
 
-    if (stat_struct.st_size > 2147483647)
+    if (stat_struct.st_size > 1099511627775)
     {
-	fprintf(stderr, "Cannot handle files larger than 2GB at this time\n");
+	fprintf(stderr, "Cannot handle files larger than 1 terabyte at this time\n");
 	return ERROR_OCCURRED;
     }
 
@@ -1386,9 +1625,9 @@ CheckFileResult nwos_check_file_checksums(char* path)
 #endif
 	nwos_read_object_from_disk(&file_ref, &file_obj, sizeof(file_obj));  /* read the file object */
 
-	file_length = (file_obj.size[0] << 24) |(file_obj.size[1] << 16) | (file_obj.size[2] << 8) | file_obj.size[3];
+	file_length = file_size_to_uint64(file_obj.size);
 
-	printf("file size: %u\n", file_length);
+	printf("file size: %llu\n", file_length);
 	/* now open the file and start creating the data block objects */
 
 	fp = fopen(path, "rb");
@@ -1425,7 +1664,7 @@ CheckFileResult nwos_check_file_checksums(char* path)
 
 	fclose(fp);
 
-	printf("length: %u\n", file_length);
+	printf("length: %llu\n", file_length);
 
 	MD5Final(md5_digest, &md5_context);   /* finish computing the md5 sum */
 	sha1_finish_ctx(&sha1_context, sha1_digest);  /* and the sha1 sum */
@@ -1433,7 +1672,7 @@ CheckFileResult nwos_check_file_checksums(char* path)
 
 	fflush(stdout);
 
-	if (file_length != ((file_obj.size[0] << 24) | (file_obj.size[1] << 16) | (file_obj.size[2] << 8) | file_obj.size[3]))
+	if (file_length != file_size_to_uint64(file_obj.size))
 	{
 	    printf("file size mismatch\n");
 	}
@@ -1514,7 +1753,7 @@ bool nwos_restore_file(ObjRef* assoc_ref, char* path)
     bool result = true;
     FILE* fp;
     int i;
-    uint32 file_length;
+    uint64 file_length;
     uint32 num_blocks;
     C_struct_File file_obj;
     C_struct_Path_And_File_Association assoc_obj;
@@ -1536,9 +1775,9 @@ bool nwos_restore_file(ObjRef* assoc_ref, char* path)
 
     nwos_read_object_from_disk(&assoc_obj.file, &file_obj, sizeof(file_obj));  /* read the file object */
 
-    file_length = (file_obj.size[0] << 24) |(file_obj.size[1] << 16) | (file_obj.size[2] << 8) | file_obj.size[3];
+    file_length = file_size_to_uint64(file_obj.size);
 
-    printf("file size: %u\n", file_length);
+    printf("file size: %llu\n", file_length);
 
     ref_list_ptr = nwos_malloc_reference_list(&file_obj.block_list);
 
@@ -1685,13 +1924,13 @@ bool nwos_restore_file(ObjRef* assoc_ref, char* path)
 
 /* Compare file - does not verify checksums (just compares the data) */
 
-bool nwos_file_is_identical(ObjRef* assoc_ref, char* path)
+bool nwos_file_is_identical(ObjRef* assoc_ref, char* path, MatchCode match)
 {
     bool result = true;
     FILE* fp;
     int i;
     struct stat stat_struct;
-    uint32 file_length;
+    uint64 file_length;
     uint32 num_blocks;
     C_struct_File file_obj;
     C_struct_Path_And_File_Association assoc_obj;
@@ -1714,9 +1953,9 @@ bool nwos_file_is_identical(ObjRef* assoc_ref, char* path)
 
     nwos_read_object_from_disk(&assoc_obj.file, &file_obj, sizeof(file_obj));  /* read the file object */
 
-    file_length = (file_obj.size[0] << 24) |(file_obj.size[1] << 16) | (file_obj.size[2] << 8) | file_obj.size[3];
+    file_length = file_size_to_uint64(file_obj.size);
 
-    printf("file size: %u\n", file_length);
+    printf("file size: %llu\n", file_length);
 
     if (stat_struct.st_size != file_length)
     {
@@ -1810,7 +2049,7 @@ bool nwos_file_is_identical(ObjRef* assoc_ref, char* path)
 	    }
 	}
 
-	if (result)   /* the file was ok, compare it's modification time */
+	if (result && match != IgnoreTime)   /* the file was ok, compare it's modification time */
 	{
 	    nwos_convert_time_stamp_to_timeval(assoc_obj.modification_time, &tv);
 
@@ -2026,7 +2265,7 @@ bool nwos_find_matching_path_and_file_association(char* dir, char* file_name, Ob
 {
     size_t length;
     struct stat stat_struct;
-    uint32 file_length;
+    uint64 file_length;
     uint8 md5_digest[16];
     uint8 sha1_digest[20];
     ObjRef path_ref;
@@ -2055,19 +2294,19 @@ bool nwos_find_matching_path_and_file_association(char* dir, char* file_name, Ob
     if (stat(path, &stat_struct) != 0)
     {
 	perror(path);
-	return ERROR_OCCURRED;       /* FIX THIS to handle this more gracefully */
+	return false;       /* FIX THIS to handle this more gracefully */
     }
 
     if (stat_struct.st_size == 0)
     {
 	fprintf(stderr, "Cannot handle empty files at this time\n");
-	return ERROR_OCCURRED;
+	return false;
     }
 
-    if (stat_struct.st_size > 2147483647)
+    if (stat_struct.st_size > 1099511627775)
     {
-	fprintf(stderr, "Cannot handle files larger than 2GB at this time\n");
-	return ERROR_OCCURRED;
+	fprintf(stderr, "Cannot handle files larger than 1 terabyte at this time\n");
+	return false;
     }
 
 
@@ -2407,7 +2646,15 @@ bool nwos_find_disc_copy(ObjRef* disc_list, int copy_num, ObjRef* ref)
 
 	    if (is_same_object(&disc_obj.disc_list, disc_list) && disc_obj.copy_number == copy_num)   /* found a match */
 	    {
-		memcpy(ref, &ref_list->references[i], sizeof(ObjRef));
+		/* if there is a newer version get it */
+		while (!is_void_reference(&disc_obj.header.object.next_version))
+		{
+		    nwos_read_object_from_disk(&disc_obj.header.object.next_version, &disc_obj, sizeof(disc_obj));
+		    assert(is_same_object(&disc_obj.disc_list, disc_list));
+		    assert(disc_obj.copy_number == copy_num);
+		}
+
+		copy_reference(ref, &disc_obj.header.common.id);
 		break;
 	    }
 	}
@@ -2433,7 +2680,62 @@ ObjCreateResult nwos_create_disc_copy(ObjRef* disc_list, int copy_num, ObjRef* l
 
     /* first find out if we already have this disc copy */
 
-    if (!nwos_find_disc_copy(disc_list, copy_num, ref))   /* didn't find it */
+    if (nwos_find_disc_copy(disc_list, copy_num, ref))   /* already exists */
+    {
+	nwos_read_object_from_disk(ref, &disc_copy_obj, sizeof(disc_copy_obj));
+
+	if (is_same_object(&disc_copy_obj.location, location))
+	{
+	    /* add verification here */
+	}
+	else  /* location changed, create new object */
+	{
+	    nwos_find_or_create_private_class_definition("DISC COPY", &disc_class_ref);
+
+	    /* verify ref is == to object ID in header */
+	    assert(is_same_object(ref, &disc_copy_obj.header.common.id));
+	    assert(is_void_reference(&disc_copy_obj.header.object.next_version));
+
+	    nwos_generate_new_id(ref);
+
+	    copy_reference(&disc_copy_obj.header.object.next_version, ref);
+
+	    nwos_crc32_calculate((uint8*) &disc_copy_obj.header.object,
+				 sizeof(ObjectHeader),
+				 disc_copy_obj.header.common.header_chksum);
+
+	    nwos_overwrite_object_to_disk(&disc_copy_obj.header.common.id, &disc_copy_obj, sizeof(disc_copy_obj));
+
+	    /* clear out next pointer that we just wrote */
+	    void_reference(&disc_copy_obj.header.object.next_version);
+
+	    /* copy the old object into the previous pointer */
+	    copy_reference(&disc_copy_obj.header.object.prev_version, &disc_copy_obj.header.common.id);
+
+	    /* put the new id in */
+	    copy_reference(&disc_copy_obj.header.common.id, ref);
+
+	    /* change the location */
+	    copy_reference(&disc_copy_obj.location, location);
+
+	    nwos_crc32_calculate((uint8*) &disc_copy_obj.header.object,
+				 sizeof(ObjectHeader),
+				 disc_copy_obj.header.common.header_chksum);
+
+	    nwos_crc32_calculate((uint8*) &disc_copy_obj.disc_list,
+				 sizeof(disc_copy_obj) - sizeof(EveryObject),
+				 disc_copy_obj.header.common.data_chksum);
+
+	    nwos_write_object_to_disk(ref, &disc_copy_obj, sizeof(disc_copy_obj));
+
+	    nwos_add_to_references(ref, &disc_class_ref);
+	    nwos_add_to_references(ref, &disc_copy_obj.disc_list);
+
+
+	    nwos_add_to_references(ref, location);
+	}
+    }
+    else   /* didn't find it */
     {
 	nwos_find_or_create_private_class_definition("DISC COPY", &disc_class_ref);
 
@@ -2592,4 +2894,37 @@ ObjCreateResult nwos_create_storage_location(char* location, ObjRef* ref)
 }
 
 
+void nwos_update_file_001_object_to_current(void* object, size_t size)
+{
+    C_struct_File_001 file_001_obj;
+    C_struct_File* file_obj_ptr = (C_struct_File*) object;
+    ObjRef file_class_ref;
+
+    memcpy(&file_001_obj, object, sizeof(file_001_obj));  /* save the 001 data */
+
+#ifdef PUBLIC_MODE
+    assert(nwos_find_public_class_definition("FILE", &file_obj_ptr->header.common.class_definition));
+#else
+    assert(nwos_find_private_class_definition("FILE", &file_class_ref));
+
+    /* if a the newer class exists, change the class definition to the latest */
+    if (!is_same_object(&file_class_ref, &file_obj_ptr->header.common.class_definition))
+    {
+	copy_reference(&file_obj_ptr->header.common.class_definition, &file_class_ref);
+    }
+#endif
+
+    file_obj_ptr->size[0] = 0;  /* most significant byte didn't exist before */
+    file_obj_ptr->size[1] = file_001_obj.size[0];
+    file_obj_ptr->size[2] = file_001_obj.size[1];
+    file_obj_ptr->size[3] = file_001_obj.size[2];
+    file_obj_ptr->size[4] = file_001_obj.size[3];
+
+    copy_reference(&file_obj_ptr->md5sum,     &file_001_obj.md5sum);
+    copy_reference(&file_obj_ptr->sha1sum,    &file_001_obj.sha1sum);
+    void_reference(&file_obj_ptr->sha256sum);
+    copy_reference(&file_obj_ptr->media,      &file_001_obj.media);
+    copy_reference(&file_obj_ptr->block_list, &file_001_obj.block_list);
+}
+
 
